#!/usr/bin/perl -w
#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: server.cgi,v 1.28 2004/08/25 01:00:47 fuktommy Exp $
#

use lib qw(..);
use Socket;
use Shingetsu::Config;
use Shingetsu::Node;
use Shingetsu::NodeList;
use Shingetsu::Cache;
use Shingetsu::Util;

umask $Shingetsu::Config::umask;

print "Content-type: text/plain\r\n",
      "X-Shingetsu: shinGETsu/$Shingetsu::Config::version\r\n\r\n";

if ((defined $ENV{PATH_INFO}) && ($ENV{PATH_INFO} ne "")) {
	$path = $ENV{PATH_INFO};
	$path =~ s|^/||;
} elsif ((defined $ENV{QUERY_STRING}) && ($ENV{QUERY_STRING} ne "")) {
	$path = $ENV{QUERY_STRING};
} else {
	$path = "";
}

if (($ENV{REQUEST_METHOD} ne "GET") && ($ENV{REQUEST_METHOD} ne "HEAD")) {
	exit;

} elsif (($path eq "") && open(IN, $Shingetsu::Config::motd)) {
	print while (<IN>);
	close IN;

} elsif ($path eq "ping") {
	print "PONG\n", "$ENV{REMOTE_ADDR}\n";

} elsif ($path eq "node") {
	my $node = Shingetsu::NodeList::random();
	if ($node) {
		$node = $node->toString();
		print $node, "\n";
	}

} elsif ($path =~ m|^join/([^:]*):(\d+)(.*)|) {
	my($host, $nport, $path) = ($1, $2, $3);
	$host = $ENV{REMOTE_ADDR} if ($host eq "");
	my $node = new Shingetsu::Node($host, $nport, $path);
	unless (checkAddress($host,$ENV{REMOTE_ADDR})&&($node->ping())) {
	} elsif (Shingetsu::NodeList::include($node) ||
		(Shingetsu::NodeList::count() < $Shingetsu::Config::nodes)) {
		Shingetsu::NodeList::add($node);
		print "WELCOME\n";
	} else {
		my $suggest = Shingetsu::NodeList::random();
		Shingetsu::NodeList::remove($suggest);
		$suggest->bye();
		print "WELCOME\n",
		      $suggest->toString(), "\n";
		Shingetsu::NodeList::add($node);
	}

} elsif ($path =~ m|^bye/([^:]*):(\d+)(.*)|) {
	my($host, $port, $path) = ($1, $2, $3);
	$host = $ENV{REMOTE_ADDR} if ($host eq "");
	my $node = new Shingetsu::Node($host, $port, $path);
	if (checkAddress($host,$ENV{REMOTE_ADDR})) {
		Shingetsu::NodeList::remove($node);
		print "BYEBYE\n";
	}

} elsif ($path =~ m|^have/([0-9A-Za-z_]+)$|) {
	my $file = $1;
	my $datadir = $Shingetsu::Config::datadir;
	if (-s "$datadir/$file.dat") {
		print "YES\n";
	} else {
		print "NO\n";
	}

} elsif ($path =~ m|^get/([0-9A-Za-z_]+)/([0-9]+)/([0-9a-f]+)$|) {
	my($file, $stamp, $id) = ($1, $2, $3);
	if (open IN, "$Shingetsu::Config::cachedir/$file/record/${stamp}_${id}.cache") {
		while (<IN>) {
			print $_, "\n";
		}
	}

} elsif ($path =~ /^(get|head)\/([0-9A-Za-z_]+)\/([-0-9A-Za-z\/]*)$/) {
	my($command, $file, $stamp) = ($1, $2, $3);
	my $datafile;
	if ($command eq "get") {
		$datafile = "$Shingetsu::Config::datadir/$file.dat";
	} else {
		$datafile = "$Shingetsu::Config::cachedir/$file/head.cache";
	}
	if (open IN, $datafile) {
		my @data = <IN>;
		close IN;
		my($begin, $end ,$id);
		local @_ = split /<>/, $data[-1];
		my $laststamp = $_[0];
		if ($stamp eq "" ) {
			($begin, $end) = (0, $laststamp);
		} elsif ($stamp =~ /^(\d+)$/) {
			($begin, $end) = ($1, $1);
		} elsif ($stamp =~ /^-(\d+)$/) {
			($begin, $end) = (0, $1);
		} elsif ($stamp =~ /^(\d+)-$/) {
			($begin, $end) = ($1, $laststamp);
		} elsif ($stamp =~ /^(\d+)-(\d+)$/) {
			($begin, $end) = ($1, $2);
		} elsif ($stamp =~ m|^(\d+)/([0-9A-Za-z]+)|) {
			($begin, $end, $id) = ($1, $1, $2);
		}
		local $_;
		if (defined $begin) {
			foreach (@data) {
				@_ = split /<>/;
				next if (@_ <= 2);
				if (($begin<=$_[0])&&($_[0]<=$end)) {
					if (defined $id) {
						print if ($_[1] eq $id);
					} else {
						print;
					}
				}
			}
		}
	}

} elsif ($path =~ m|^update/([0-9A-Za-z_]+)/(\d+)/([0-9A-Za-z]+)/([^:]*):(\d+)(.*)|) {
	my ($file, $stamp, $id, $host, $nport, $path) = ($1, $2, $3, $4, $5, $6);
	$host = $ENV{REMOTE_ADDR} if ($host eq "");
	$node = new Shingetsu::Node($host, $nport, $path);
	update($file, $stamp, $id, $node) and print "OK\n";
}

#
# relay update infomation
#
sub update {
	my($file, $stamp, $id, $node) = @_;
	my $datadir     = $Shingetsu::Config::datadir;
	my $updatelist  = $Shingetsu::Config::updatelist;
	local $_;
	if ($stamp < time - $Shingetsu::Config::saveupdate * 60) {
		return 0;
	}
	my $idline = "$stamp<>$id<>$file\n";
	my @update = ();
	if (open IN, $updatelist) {
		@update = <IN>;
		close IN;
	}
	if (grep $_ eq $idline, @update) {
		return 1;
	}
	Shingetsu::Util::addUpdate($stamp, $id, $file);
	my $flag = 0;
	if (! (-f "$datadir/$file.dat")) {
		Shingetsu::Cache::tellupdate($file, $stamp, $id, $node);
		return;
	} elsif (-s "$datadir/$file.dat") {
		$flag = Shingetsu::Cache::getData($file, $stamp, $id, $node);
	} else {
		Shingetsu::Cache::getRegion($file, $node);
		$flag = (-s "$datadir/$file.dat");
	}
	if ($flag) {
		Shingetsu::Cache::tellupdate($file, $stamp, $id, undef);
		unless (Shingetsu::NodeList::include($node)
		     ||(Shingetsu::NodeList::count() > $Shingetsu::Config::nodes)) {
			Shingetsu::NodeList::join($node);
		}
		return 1;
	} else {
		removeUpdate($idline);
		return 0;
	}
}

#
# check client == node ?
#
sub checkAddress {
	my($host, $ip) = @_;
	return 1 if ($host eq $ip);
	my $i;
	for ($i=0; $i<=$Shingetsu::Config::retry; $i++) {
		my $addr;
		if ($addr = inet_ntoa inet_aton $host) {
			return $addr eq $ip;
		}
	}
	return 0;
}

#
# remove from update list
#
sub removeUpdate {
	my($idline) = @_;
	my $updatelist  = $Shingetsu::Config::updatelist;
	my @update = ();
	if (open IN, $updatelist) {
		@update = <IN>;
		close IN;
	}
	Shingetsu::Util::lock();
	open OUT, "> $updatelist" or die "$updatelist: failed to open. $!";
	foreach (@update) {
		print OUT if ($_ ne $idline);
	}
	close OUT;
	Shingetsu::Util::unlock();
}
